/*
	written on mySQL 4.x
	by Louis Solomon / www.SteelBytes.com
	23/Oct/2003 - 04/Nov/2003
*/

#if defined(_DEBUG) || 0
#define MY_TRACE_ON
#endif

#pragma warning (disable:4100) // unreferenced formal parameter
#pragma warning (disable:4127) // conditional expression is constant
#pragma warning (disable:4164) // intrinsic function not declared
#pragma warning (disable:4201) // nonstandard extension used : nameless struct/union
#pragma warning (disable:4206) // nonstandard extension used : translation unit is empty
#pragma warning (disable:4514) // unreferenced inline function has been removed

#define STRSAFE_LIB

#define STRICT
#define VC_EXTRALEAN
#define WIN32_LEAN_AND_MEAN

//#if defined(_PREFAST_)
// because of MS's fuck ups causing CRT to clash with PlatformSDK in VS2005Beta2
//#define _FORCE_SPECSTRINGS_DECLSPECS
//#include "C:\Program Files\Microsoft Platform SDK\Include\specstrings.h"
//#endif

/*lint -save -e* */
#include <windows.h>
#include <sql.h>
#include <sqlext.h>
#include <stdlib.h>
#include <strsafe.h>
/*lint -restore */;
#include <stdlib.h>

#pragma hdrstop

#ifndef _DEBUG
#pragma check_stack(off)
#pragma runtime_checks( "", off )
#endif

#pragma intrinsic(memset)
#pragma intrinsic(memcmp)
#pragma intrinsic(memcpy)
//#pragma intrinsic(strcpy)
#pragma intrinsic(strcmp)
#ifdef __cplusplus
extern "C" {
#endif
#undef RtlCopyMemory
#undef RtlZeroMemory
#undef RtlMoveMemory
NTSYSAPI VOID NTAPI RtlMoveMemory (VOID UNALIGNED *Destination,CONST VOID UNALIGNED *Source,SIZE_T Length);
NTSYSAPI VOID NTAPI RtlZeroMemory (VOID UNALIGNED *Destination,SIZE_T Length);
#define RtlCopyMemory RtlMoveMemory
#ifdef __cplusplus
}
#endif

#define STRING_LENGTH(s) (sizeof(s)/sizeof((s)[0]))
#include "ptinterface.h"

#pragma comment(linker,"/EXPORT:PTI_HttpLog_Start=_PTI_HttpLog_Start@4")
#pragma comment(linker,"/EXPORT:PTI_HttpLog_End=_PTI_HttpLog_End@4")
#pragma comment(linker,"/EXPORT:PTI_HttpLog_Hit=_PTI_HttpLog_Hit@4")

static SQLHENV henv;
static SQLHDBC hdbc1;
static LONG refcount;
static char filename_ini[MAX_PATH];

#define single_string_val_size 1024
typedef char char1024[single_string_val_size];

static void init_string_lengths(unsigned num,const char * const *strings,int *lengths)
{
	for (unsigned i=0;i<num;++i)
	{
		lengths[i] = lstrlen(strings[i]);
	}
}

static void mysql_encode(char *string_out,int string_out_cc,const char *string_in)
{
	const char *p = string_in;
	char *q = string_out;
	int q_avail = string_out_cc-1; // allow room for null term
	while (*p)
	{
		if (q_avail<=0)
			goto end;
		switch (*p)
		{
		case 26:
			if (q_avail<2)
				goto end;
			*q++ = '\\';
			*q++ = 'z';
			q_avail-=2;
			break;
		case '\r':
			if (q_avail<2)
				goto end;
			*q++ = '\\';
			*q++ = 'r';
			q_avail-=2;
			break;
		case '\n':
			if (q_avail<2)
				goto end;
			*q++ = '\\';
			*q++ = 'n';
			q_avail-=2;
			break;
		case '\t':
			if (q_avail<2)
				goto end;
			*q++ = '\\';
			*q++ = 't';
			q_avail-=2;
			break;
		case '\b':
			if (q_avail<2)
				goto end;
			*q++ = '\\';
			*q++ = 'b';
			q_avail-=2;
			break;
		case '\'':
		case '\"':
		case '\\':
			if (q_avail<2)
				goto end;
			*q++ = '\\';
			*q++ = *p++;
			q_avail-=2;
			break;
		default:
			*q++ = *p++;
			q_avail--;
			break;
		}
	}
end:
	*q = 0;
}

static void replace_tokens(char *string_out,int string_out_cc,const char *string_in,int strings_num,const char * const *strings_tags,const int *strings_tags_lengths,const char (*strings_vals)[single_string_val_size])
{
	const char *p = string_in;
	char *q = string_out;
	int q_avail = string_out_cc-1; // allow room for null term
	char *temp = (char *)malloc(1024);
	if (temp)
	{
	while (*p)
	{
		if (q_avail<=0)
			break;
		int i;
		for (i = 0; i<strings_num; ++i)
		{
			if (memcmp(p,strings_tags[i],(size_t)strings_tags_lengths[i])==0)
			{
				p += strings_tags_lengths[i];
				mysql_encode(temp,1024,strings_vals[i]);
/*#ifdef MY_TRACE_ON
				{
					char s[4096];
					StringCchPrintf(s,STRING_LENGTH(s),"replacing '%s' with '%s'\r\n",strings_tags[i],temp);
					OutputDebugString(s);
				}
#endif*/
				//int old_q_avail = q_avail;
				//StringCchCopyEx(q,(size_t)q_avail,temp,NULL,&q_avail,0);
				//q += old_q_avail-q_avail;

				int l = lstrlen(temp);
				int m = min(l,q_avail);
				memcpy(q,temp,(size_t)m);
				q += m;
				q_avail -= m;
				break;
			}
		}
		if (i==strings_num)
		{
			*q++ = *p++;
			q_avail--;
		}
	}
	free(temp);
	}
	*q = 0;
}

static bool SQLisOK(SQLRETURN code)
{
	return (code==SQL_SUCCESS || code==SQL_SUCCESS_WITH_INFO);
}

static void DB_done(void)
{
	if (InterlockedDecrement(&refcount)==0)
	{
		if (henv!=SQL_NULL_HENV)
		{
			if (hdbc1!=SQL_NULL_HDBC)
			{
				SQLDisconnect(hdbc1);
				SQLFreeHandle(SQL_HANDLE_DBC, hdbc1);
				hdbc1 = SQL_NULL_HDBC;
			}
			SQLFreeHandle(SQL_HANDLE_ENV, henv);
			henv = SQL_NULL_HENV;
		}
	}
}

static void DB_init(void)
{
	InterlockedIncrement(&refcount);
	if (SQLisOK(SQLAllocHandle(SQL_HANDLE_ENV, NULL, &henv)) && SQLisOK(SQLSetEnvAttr(henv, SQL_ATTR_ODBC_VERSION,(SQLPOINTER) SQL_OV_ODBC3, SQL_IS_INTEGER)))
	{
		if (SQLisOK(SQLAllocHandle(SQL_HANDLE_DBC, henv, &hdbc1)))
		{
			char dsn_name[256];
			char dsn_user[256];
			char dsn_pass[256];
			GetPrivateProfileString("ODBC-DSN","Name","",dsn_name,STRING_LENGTH(dsn_name),filename_ini);
			GetPrivateProfileString("ODBC-DSN","User","",dsn_user,STRING_LENGTH(dsn_user),filename_ini);
			GetPrivateProfileString("ODBC-DSN","Pass","",dsn_pass,STRING_LENGTH(dsn_pass),filename_ini);
			if (SQLisOK(SQLConnect(hdbc1,(SQLCHAR*)dsn_name, SQL_NTS, (SQLCHAR*)dsn_user, SQL_NTS,(SQLCHAR*)dsn_pass, SQL_NTS)))
			{
				// init ok.
			}
			else
			{
				DB_done();
			}
		}
	}
}

static bool DB_DoQuery(SQLHSTMT hstmt1,const char *query,SQLSMALLINT *num_results_returned=NULL,SQLINTEGER *data=NULL)
{
	bool result = false;
	if (num_results_returned)
		*num_results_returned = 0;
	if (SQLisOK(SQLExecDirect(hstmt1,(SQLCHAR*)query,SQL_NTS)))
	{
		if (num_results_returned)
		{
			SQLNumResultCols(hstmt1,num_results_returned);
			if (((*num_results_returned)>0) && (data!=NULL))
			{
				if (SQLisOK(SQLFetch(hstmt1)))
				{
					SQLGetData(hstmt1,1,SQL_C_SLONG,data,sizeof(*data),NULL);
				}
			}
		}
		SQLCloseCursor(hstmt1);
		result = true;
/*#ifdef MY_TRACE_ON
		{
			char s[4096];
			StringCchPrintf(s,STRING_LENGTH(s),"The following SQL succeded: %s\r\n",query);	
			OutputDebugString(s);
		}
#endif*/
	}
	else
	{
#ifdef MY_TRACE_ON
		{
			char s[4096];
			StringCchPrintf(s,STRING_LENGTH(s),"The following SQL failed: %s\r\n",query);	
			OutputDebugString(s);
		}
#endif
	}
	return result;
}

/*static void GetFirstParam(const char *params,char *firstparam,unsigned firstparam_cc)
{
	const char *p = params;
	while (*p && *p!=' ' && *p!='\t' && *p!='\r' && *p!='\n')
		p++;
	unsigned l = p-params;
	l = min(firstparam_cc-1,l);
	memcpy(firstparam,params,l);
	firstparam[l] = 0;
}*/

typedef struct {
	WORD portin,portin2;
	char tablecreate[4096];
	char tableinsert[4096];
	//char param1[256];
} USERDATA;

static const char *tablecreate_strings_tags[] = {
	"<pti_mappingname>",
	//"<pti_param1>",
};
static int tablecreate_strings_tags_lengths[_countof(tablecreate_strings_tags)];
void PTI_API PTI_HttpLog_Start(PTI_HTTPLOG_START *httplog_start)
{
#ifdef MY_TRACE_ON
	OutputDebugString("pti_httplog3::PTI_HttpLog_Start() - start\r\n");
#endif
	httplog_start->userdata = (void *)malloc(sizeof(USERDATA));
	if (httplog_start->userdata)
	{
		((USERDATA *)httplog_start->userdata)->portin = httplog_start->portin;
		((USERDATA *)httplog_start->userdata)->portin2 = httplog_start->portin2;
		//GetFirstParam(httplog_start->params,((USERDATA *)httplog_start->userdata)->param1,_countof(((USERDATA *)httplog_start->userdata)->param1));
		char temp[256];
		StringCchPrintf(temp,STRING_LENGTH(temp),"SQL%s",httplog_start->params);
		GetPrivateProfileString(temp,"TableCreate","",((USERDATA *)httplog_start->userdata)->tablecreate,STRING_LENGTH(((USERDATA *)httplog_start->userdata)->tablecreate),filename_ini);
		GetPrivateProfileString(temp,"TableInsert","",((USERDATA *)httplog_start->userdata)->tableinsert,STRING_LENGTH(((USERDATA *)httplog_start->userdata)->tableinsert),filename_ini);
	#ifdef MY_TRACE_ON
		{
			char s[4096];
			StringCchPrintf(s,STRING_LENGTH(s),"create=%s\r\ninsert=%s\r\n",((const USERDATA *)httplog_start->userdata)->tablecreate,((const USERDATA *)httplog_start->userdata)->tableinsert);	
			OutputDebugString(s);
		}
	#endif
		DB_init();
		if (hdbc1!=SQL_NULL_HDBC)
		{
			SQLHSTMT hstmt1 = SQL_NULL_HSTMT;
			if (SQLisOK(SQLAllocHandle(SQL_HANDLE_STMT, hdbc1, &hstmt1)))
			{
				char1024 *strings_vals = (char1024 *)malloc(_countof(tablecreate_strings_tags)*sizeof(char1024));
				if (strings_vals)
				{
					StringCchPrintf(strings_vals[0],STRING_LENGTH(strings_vals[0]),"%s",httplog_start->mapping_name);
					//lstrcpyn(strings_vals[1],((const USERDATA *)httplog_start->userdata)->param1,_countof(strings_vals[1]));
					char *query = (char *)malloc(4096);
					if (query)
					{
						replace_tokens(query,4096,((USERDATA *)httplog_start->userdata)->tablecreate,_countof(tablecreate_strings_tags),tablecreate_strings_tags,tablecreate_strings_tags_lengths,strings_vals);
						DB_DoQuery(hstmt1,query);
						free(query);
					}
					free(strings_vals);
				}
				SQLFreeHandle(SQL_HANDLE_STMT, hstmt1);
			}
		}
	}
#ifdef MY_TRACE_ON
	OutputDebugString("pti_httplog3::PTI_HttpLog_Start() - end\r\n");
#endif
}

static const char *tableinsert_strings_tags[] = {
	"<pti_mappingname>",
	"<pti_request_line>",
	"<pti_reply_line>",
	"<pti_header_host>",
	"<pti_client_ip>",
	"<pti_length_in>",
	"<pti_length_out>",
	//"<pti_param1>",
};
static int tableinsert_strings_tags_lengths[_countof(tableinsert_strings_tags)];
/*lint -e{818} */
int PTI_API PTI_HttpLog_Hit(PTI_HTTPLOG_HIT *httplog_hit)
{
	int res = -1;
#ifdef MY_TRACE_ON
	OutputDebugString("pti_httplog3::PTI_HttpLog_Hit() - start\r\n");
#endif
	if (hdbc1!=SQL_NULL_HDBC)
	{
		SQLHSTMT hstmt1 = SQL_NULL_HSTMT;
		if (SQLisOK(SQLAllocHandle(SQL_HANDLE_STMT, hdbc1, &hstmt1)))
		{
			char1024 *strings_vals = (char1024 *)malloc(_countof(tableinsert_strings_tags)*sizeof(char1024));
			if (strings_vals)
			{
			StringCchPrintf(strings_vals[0],STRING_LENGTH(strings_vals[0]),"%s",httplog_hit->mapping_name);
			StringCchPrintf(strings_vals[1],STRING_LENGTH(strings_vals[1]),"%s",httplog_hit->request_line);
			StringCchPrintf(strings_vals[2],STRING_LENGTH(strings_vals[2]),"%s",httplog_hit->reply_line);
			StringCchPrintf(strings_vals[3],STRING_LENGTH(strings_vals[3]),"%s",httplog_hit->header_host);
			CharLower(strings_vals[3]);
			StringCchPrintf(strings_vals[4],STRING_LENGTH(strings_vals[4]),"%i.%i.%i.%i",((BYTE*)&httplog_hit->client_ip)[3],((BYTE*)&httplog_hit->client_ip)[2],((BYTE*)&httplog_hit->client_ip)[1],((BYTE*)&httplog_hit->client_ip)[0]);
			StringCchPrintf(strings_vals[5],STRING_LENGTH(strings_vals[5]),"%i",httplog_hit->length_in);
			StringCchPrintf(strings_vals[6],STRING_LENGTH(strings_vals[6]),"%i",httplog_hit->length_out);
			//lstrcpyn(strings_vals[7],((const USERDATA *)httplog_hit->userdata)->param1,_countof(strings_vals[7]));

			{
				// strip trailing :80
				char *p = strings_vals[3] + lstrlen(strings_vals[3]);
				while (p>=strings_vals[3])
				{
					if (*p==':')
					{
						WORD prt = (WORD)atoi(p+1);
						if (prt==80 || prt==443) // kludgy, as it should only check which ever ... probably should use a pti param :-)
							//if (prt>=((USERDATA *)httplog_hit->userdata)->portin && prt <=((USERDATA *)httplog_hit->userdata)->portin2)
						{
							*p = 0;
						}
						break;
					}
					else
					{
						p--;
					}
				}
			}

			char *query = (char *)malloc(4096);
			if (query)
			{
				replace_tokens(query,4096,((USERDATA *)httplog_hit->userdata)->tableinsert,_countof(tableinsert_strings_tags),tableinsert_strings_tags,tableinsert_strings_tags_lengths,strings_vals);
				if (DB_DoQuery(hstmt1,query))
				{
					res = 0;
				}
				free(query);
			}
			free(strings_vals);
			}
			SQLFreeHandle(SQL_HANDLE_STMT, hstmt1);
		}
	}
#ifdef MY_TRACE_ON
	OutputDebugString("pti_httplog3::PTI_HttpLog_Hit() - end\r\n");
#endif
	return res;
}

void PTI_API PTI_HttpLog_End(PTI_HTTPLOG_END *httplog_end)
{
#ifdef MY_TRACE_ON
	OutputDebugString("pti_httplog3::PTI_HttpLog_End() - start\r\n");
#endif
	DB_done();
	free(((USERDATA *)httplog_end->userdata));
	httplog_end->userdata = NULL;
#ifdef MY_TRACE_ON
	OutputDebugString("pti_httplog3::PTI_HttpLog_End() - end\r\n");
#endif
}

/*lint -e{715,818} */
BOOL WINAPI DllMain(HINSTANCE hinstDLL,DWORD fdwReason,LPVOID lpvReserved)
{
#ifdef MY_TRACE_ON
	OutputDebugString("pti_httplog3::DllMain() - start\r\n");
#endif
	if (fdwReason==DLL_PROCESS_ATTACH)
	{
		init_string_lengths(_countof(tablecreate_strings_tags),tablecreate_strings_tags,tablecreate_strings_tags_lengths);
		init_string_lengths(_countof(tableinsert_strings_tags),tableinsert_strings_tags,tableinsert_strings_tags_lengths);
		GetModuleFileName(hinstDLL,filename_ini,STRING_LENGTH(filename_ini));
		char *p = filename_ini+lstrlen(filename_ini);
		while (p>filename_ini && *p!='.')
			p--;
		*p = 0;
		StringCchCat(filename_ini,STRING_LENGTH(filename_ini),".INI");
	}
#ifdef MY_TRACE_ON
	OutputDebugString("pti_httplog3::DllMain() - end\r\n");
#endif
	return 1;
}
