/*
	written on mySQL 4.x
	by Louis Solomon / www.SteelBytes.com
	24/Nov/03
*/

#pragma warning (disable:4100) // unreferenced formal parameter
#pragma warning (disable:4127) // conditional expression is constant
#pragma warning (disable:4164) // intrinsic function not declared
#pragma warning (disable:4201) // nonstandard extension used : nameless struct/union
#pragma warning (disable:4206) // nonstandard extension used : translation unit is empty
#pragma warning (disable:4514) // unreferenced inline function has been removed

#define STRSAFE_LIB

#define STRICT
#define VC_EXTRALEAN
#define WIN32_LEAN_AND_MEAN

//#if defined(_PREFAST_)
//// because of MS's fuck ups causing CRT to clash with PlatformSDK in VS2005Beta2
//#define _FORCE_SPECSTRINGS_DECLSPECS
//#include "C:\Program Files\Microsoft Platform SDK\Include\specstrings.h"
//#endif

/*lint -save -e* */
#include <windows.h>
#include <sql.h>
#include <sqlext.h>
//#include <odbcss.h>
#include <strsafe.h>
/*lint -restore */
#include <stdlib.h>

#pragma hdrstop

#ifndef _DEBUG
#pragma check_stack(off)
#pragma runtime_checks( "", off )
#endif

#pragma intrinsic(memset)
#pragma intrinsic(memcmp)
#pragma intrinsic(memcpy)
//#pragma intrinsic(strcpy)
#pragma intrinsic(strcmp)
#ifdef __cplusplus
extern "C" {
#endif
#undef RtlCopyMemory
#undef RtlZeroMemory
#undef RtlMoveMemory
NTSYSAPI VOID NTAPI RtlMoveMemory (VOID UNALIGNED *Destination,CONST VOID UNALIGNED *Source,SIZE_T Length);
NTSYSAPI VOID NTAPI RtlZeroMemory (VOID UNALIGNED *Destination,SIZE_T Length);
#define RtlCopyMemory RtlMoveMemory
#ifdef __cplusplus
}
#endif

#define STRING_LENGTH(s) (sizeof(s)/sizeof((s)[0]))
#include "ptinterface.h"

#pragma comment(linker,"/EXPORT:PTI_StatsEvent_Start=_PTI_StatsEvent_Start@4")
#pragma comment(linker,"/EXPORT:PTI_StatsEvent_End=_PTI_StatsEvent_End@4")
#pragma comment(linker,"/EXPORT:PTI_StatsEvent_Event=_PTI_StatsEvent_Event@4")

static SQLHENV henv = SQL_NULL_HENV;
static SQLHDBC hdbc1 = SQL_NULL_HDBC;
static LONG refcount = 0;
static char filename_ini[MAX_PATH];

#define single_string_val_size 1024
typedef char char1024[single_string_val_size];

static void init_string_lengths(int num,const char *const *strings,int *lengths)
{
	for (int i=0;i<num;++i)
	{
		lengths[i] = lstrlen(strings[i]);
	}
}

static void mysql_encode(char *string_out,int string_out_cc,const char *string_in)
{
	const char *p = string_in;
	char *q = string_out;
	int q_avail = string_out_cc-1; // allow room for null term
	while (*p)
	{
		if (q_avail<=0)
			goto end;
		switch (*p)
		{
		case 26:
			if (q_avail<2)
				goto end;
			*q++ = '\\';
			*q++ = 'z';
			q_avail-=2;
			break;
		case '\r':
			if (q_avail<2)
				goto end;
			*q++ = '\\';
			*q++ = 'r';
			q_avail-=2;
			break;
		case '\n':
			if (q_avail<2)
				goto end;
			*q++ = '\\';
			*q++ = 'n';
			q_avail-=2;
			break;
		case '\t':
			if (q_avail<2)
				goto end;
			*q++ = '\\';
			*q++ = 't';
			q_avail-=2;
			break;
		case '\b':
			if (q_avail<2)
				goto end;
			*q++ = '\\';
			*q++ = 'b';
			q_avail-=2;
			break;
		case '\'':
		case '\"':
		case '\\':
			if (q_avail<2)
				goto end;
			*q++ = '\\';
			*q++ = *p++;
			q_avail-=2;
			break;
		default:
			*q++ = *p++;
			q_avail--;
			break;
		}
	}
end:
	*q = 0;
}

static void replace_tokens(char *string_out,int string_out_cc,const char *string_in,int strings_num,const char * const *strings_tags,const int *strings_tags_lengths,const char (*strings_vals)[single_string_val_size])
{
	const char *p = string_in;
	char *q = string_out;
	int q_avail = string_out_cc-1; // allow room for null term
	char *temp = (char *)malloc(1024);
	if (temp)
	{
	while (*p)
	{
		if (q_avail<=0)
			break;
		int i;
		for (i = 0; i<strings_num; ++i)
		{
			if (memcmp(p,strings_tags[i],(size_t)strings_tags_lengths[i])==0)
			{
				p += strings_tags_lengths[i];
				
				mysql_encode(temp,1024,strings_vals[i]);
/*#ifdef _DEBUG
				{
					char s[4096];
					StringCchPrintf(s,STRING_LENGTH(s),"replacing '%s' with '%s'\r\n",strings_tags[i],temp);
					OutputDebugString(s);
				}
#endif*/
				//size_t old_q_avail = q_avail;
				//StringCchCopyEx(q,q_avail,temp,NULL,&q_avail,0);
				//q += old_q_avail-q_avail;

				int l = lstrlen(temp);
				int m = min(l,q_avail);
				memcpy(q,temp,(size_t)m);
				q += m;
				q_avail -= m;

				break;
			}
		}
		if (i==strings_num)
		{
			*q++ = *p++;
			q_avail--;
		}
	}
	free(temp);
	}
	*q = 0;
}

static bool SQLisOK(SQLRETURN code)
{
	return (code==SQL_SUCCESS || code==SQL_SUCCESS_WITH_INFO);
}

static void DB_done(void)
{
	if (InterlockedDecrement(&refcount)==0)
	{
		if (henv!=SQL_NULL_HENV)
		{
			if (hdbc1!=SQL_NULL_HDBC)
			{
				SQLDisconnect(hdbc1);
				SQLFreeHandle(SQL_HANDLE_DBC, hdbc1);
				hdbc1 = SQL_NULL_HDBC;
			}
			SQLFreeHandle(SQL_HANDLE_ENV, henv);
			henv = SQL_NULL_HENV;
		}
	}
}

static void DB_init(void)
{
	if (InterlockedIncrement(&refcount)==1)
	{
		if (SQLisOK(SQLAllocHandle(SQL_HANDLE_ENV, NULL, &henv)) && SQLisOK(SQLSetEnvAttr(henv, SQL_ATTR_ODBC_VERSION,(SQLPOINTER) SQL_OV_ODBC3, SQL_IS_INTEGER)))
		{
			if (SQLisOK(SQLAllocHandle(SQL_HANDLE_DBC, henv, &hdbc1)))
			{
				char dsn_name[256];
				char dsn_user[256];
				char dsn_pass[256];
				GetPrivateProfileString("ODBC-DSN","Name","",dsn_name,STRING_LENGTH(dsn_name),filename_ini);
				GetPrivateProfileString("ODBC-DSN","User","",dsn_user,STRING_LENGTH(dsn_user),filename_ini);
				GetPrivateProfileString("ODBC-DSN","Pass","",dsn_pass,STRING_LENGTH(dsn_pass),filename_ini);
				if (SQLisOK(SQLConnect(hdbc1,(SQLCHAR*)dsn_name, SQL_NTS, (SQLCHAR*)dsn_user, SQL_NTS,(SQLCHAR*)dsn_pass, SQL_NTS)))
				{
					// init ok.
				}
				else
				{
					DB_done();
				}
			}
		}
	}
}

static bool DB_DoQuery(SQLHSTMT hstmt1,const char *query,SQLSMALLINT *num_results_returned=NULL,SQLINTEGER *data=NULL)
{
	bool result = false;
	if (num_results_returned)
		*num_results_returned = 0;
	if (SQLisOK(SQLExecDirect(hstmt1,(SQLCHAR*)query,SQL_NTS)))
	{
		if (num_results_returned)
		{
			SQLNumResultCols(hstmt1,num_results_returned);
			if (((*num_results_returned)>0) && (data!=NULL))
			{
				if (SQLisOK(SQLFetch(hstmt1)))
				{
					SQLGetData(hstmt1,1,SQL_C_SLONG,data,sizeof(*data),NULL);
				}
			}
		}
		SQLCloseCursor(hstmt1);
		result = true;
/*#ifdef _DEBUG
		{
			char s[4096];
			StringCchPrintf(s,STRING_LENGTH(s),"The following SQL succeded: %s\r\n",query);	
			OutputDebugString(s);
		}
#endif*/
	}
	else
	{
#ifdef _DEBUG
		{
			char s[4096];
			StringCchPrintf(s,STRING_LENGTH(s),"The following SQL failed: %s\r\n",query);	
			OutputDebugString(s);
		}
#endif
	}
	return result;
}

typedef struct {
	char tablecreate[4096];
	char tableinsert[4096];
	char tableconnect[4096];
	char tablebytes[4096];
} USERDATA;

static const char *tablecreate_strings_tags[] = {
	"<pti_mappingname>",
};
static int tablecreate_strings_tags_lengths[_countof(tablecreate_strings_tags)];
void PTI_API PTI_StatsEvent_Start(PTI_STATSEVENT_START *statsevent_start)
{
	statsevent_start->userdata = (void *)malloc(sizeof(USERDATA));
	if (!statsevent_start->userdata)
		return;
	char temp[256];
	StringCchPrintf(temp,STRING_LENGTH(temp),"SQL%s",statsevent_start->params);
	GetPrivateProfileString(temp,"TableCreate","",((USERDATA *)statsevent_start->userdata)->tablecreate,STRING_LENGTH(((USERDATA *)statsevent_start->userdata)->tablecreate),filename_ini);
	GetPrivateProfileString(temp,"TableInsert","",((USERDATA *)statsevent_start->userdata)->tableinsert,STRING_LENGTH(((USERDATA *)statsevent_start->userdata)->tableinsert),filename_ini);
	GetPrivateProfileString(temp,"TableConnect","",((USERDATA *)statsevent_start->userdata)->tableconnect,STRING_LENGTH(((USERDATA *)statsevent_start->userdata)->tableconnect),filename_ini);
	GetPrivateProfileString(temp,"TableBytes","",((USERDATA *)statsevent_start->userdata)->tablebytes,STRING_LENGTH(((USERDATA *)statsevent_start->userdata)->tablebytes),filename_ini);
#ifdef _DEBUG
	{
		char s[4096];
		StringCchPrintf(s,STRING_LENGTH(s),"PTI_StatsEvent_Start()\r\n");	
		OutputDebugString(s);
	}
#endif
	DB_init();
	if (hdbc1==SQL_NULL_HDBC)
		return;
	SQLHSTMT hstmt1 = SQL_NULL_HSTMT;
	if (SQLisOK(SQLAllocHandle(SQL_HANDLE_STMT, hdbc1, &hstmt1)))
	{
		char1024 *strings_vals = (char1024 *)malloc(_countof(tablecreate_strings_tags)*sizeof(char1024));
		if (strings_vals)
		{
			StringCchPrintf(strings_vals[0],STRING_LENGTH(strings_vals[0]),"%s",statsevent_start->mapping_name);

			char *query = (char *)malloc(4096);
			if (query)
			{
				replace_tokens(query,4096,((USERDATA *)statsevent_start->userdata)->tablecreate,sizeof(tablecreate_strings_tags)/sizeof(tablecreate_strings_tags[0]),tablecreate_strings_tags,tablecreate_strings_tags_lengths,strings_vals);
				DB_DoQuery(hstmt1,query);
				free(query);
			}
			free(strings_vals);
		}
		SQLFreeHandle(SQL_HANDLE_STMT, hstmt1);
	}
}

static const char *tableinsert_strings_tags[] = {
	"<pti_mappingname>",
	"<pti_client_ip>",
	"<pti_date>",
	"<pti_bytesin>",
	"<pti_bytesout>",
};
static int tableinsert_strings_tags_lengths[_countof(tableinsert_strings_tags)];
/*lint -e{818} */
void PTI_API PTI_StatsEvent_Event(PTI_STATSEVENT_EVENT *statsevent_event)
{
	if (hdbc1==SQL_NULL_HDBC)
		return;
	if (statsevent_event->event_id==PTISE_CONNECT || statsevent_event->event_id==PTISE_BYTESINOUT)
	{
		SQLHSTMT hstmt1 = SQL_NULL_HSTMT;
		if (SQLisOK(SQLAllocHandle(SQL_HANDLE_STMT, hdbc1, &hstmt1)))
		{
			char *query = (char *)malloc(4096);
			if (query)
			{
				char1024 *strings_vals = (char1024 *)malloc(_countof(tableinsert_strings_tags)*sizeof(char1024));
				if (strings_vals)
				{
					StringCchPrintf(strings_vals[0],STRING_LENGTH(strings_vals[0]),"%s",statsevent_event->mapping_name);
					StringCchPrintf(strings_vals[1],STRING_LENGTH(strings_vals[1]),"%i.%i.%i.%i",((BYTE*)&statsevent_event->clientip)[3],((BYTE*)&statsevent_event->clientip)[2],((BYTE*)&statsevent_event->clientip)[1],((BYTE*)&statsevent_event->clientip)[0]);
					GetDateFormat(0,0,NULL,"yyyy-MM-dd",strings_vals[2],STRING_LENGTH(strings_vals[2]));
					StringCchPrintf(strings_vals[3],STRING_LENGTH(strings_vals[3]),"%I64i",statsevent_event->bytesinout.bytesin);
					StringCchPrintf(strings_vals[4],STRING_LENGTH(strings_vals[4]),"%I64i",statsevent_event->bytesinout.bytesout);

					replace_tokens(query,4096,((USERDATA *)statsevent_event->userdata)->tableinsert,sizeof(tableinsert_strings_tags)/sizeof(tableinsert_strings_tags[0]),tableinsert_strings_tags,tableinsert_strings_tags_lengths,strings_vals);
					DB_DoQuery(hstmt1,query);

					if (statsevent_event->event_id==PTISE_CONNECT)
					{
						replace_tokens(query,4096,((USERDATA *)statsevent_event->userdata)->tableconnect,sizeof(tableinsert_strings_tags)/sizeof(tableinsert_strings_tags[0]),tableinsert_strings_tags,tableinsert_strings_tags_lengths,strings_vals);
						DB_DoQuery(hstmt1,query);
					}
					else if (statsevent_event->event_id==PTISE_BYTESINOUT)
					{
						replace_tokens(query,4096,((USERDATA *)statsevent_event->userdata)->tablebytes,sizeof(tableinsert_strings_tags)/sizeof(tableinsert_strings_tags[0]),tableinsert_strings_tags,tableinsert_strings_tags_lengths,strings_vals);
						DB_DoQuery(hstmt1,query);
					}
					free(strings_vals);
				}
				free(query);
			}
			SQLFreeHandle(SQL_HANDLE_STMT, hstmt1);
		}
	}
}

void PTI_API PTI_StatsEvent_End(PTI_STATSEVENT_END *statsevent_end)
{
#ifdef _DEBUG
	{
		char s[4096];
		StringCchPrintf(s,STRING_LENGTH(s),"PTI_StatsEvent_End()\r\n");	
		OutputDebugString(s);
	}
#endif
	DB_done();
	free(((USERDATA *)statsevent_end->userdata));
	statsevent_end->userdata = NULL;
}

/*lint -e{715,818} */
BOOL WINAPI DllMain(HINSTANCE hinstDLL,DWORD fdwReason,LPVOID lpvReserved)
{
	if (fdwReason==DLL_PROCESS_ATTACH)
	{
		init_string_lengths(sizeof(tablecreate_strings_tags)/sizeof(tablecreate_strings_tags[0]),tablecreate_strings_tags,tablecreate_strings_tags_lengths);
		init_string_lengths(sizeof(tableinsert_strings_tags)/sizeof(tableinsert_strings_tags[0]),tableinsert_strings_tags,tableinsert_strings_tags_lengths);
		GetModuleFileName(hinstDLL,filename_ini,STRING_LENGTH(filename_ini));
		char *p = filename_ini+lstrlen(filename_ini);
		while (p>filename_ini && *p!='.')
			p--;
		*p = 0;
		StringCchCat(filename_ini,STRING_LENGTH(filename_ini),".INI");
	}
	return 1;
}
